#include <stdio.h>
#include <SDL2/SDL.h>
#include <GLES2/gl2.h>
#include "VectorMath.h"
#include "Camera.h"

const float PI = 3.1415926535897932384626433832795;

int screenWidth = 640;
int screenHeight = 480;
SDL_Window *win;
SDL_GLContext context;
SDL_Renderer *renderer;

int         Shader[2];              // We have a vertex & a fragment shader
int         Program;                // Totalling one program
float       Angle = 0.0;            // Rotation angle of our object
float       Proj[4][4];             // Projection matrix
int         iProj, iModel, iCamera;          // Our 3 uniforms

CMatrix cmModelviewMatrix;
CCamera ccCamera;


void InitGame();
void FinishGame();

void LoadShader(char *, int);
int InitShaders(void);
void Display(void);
void DrawTourus(float , float , int , int );
void DrawCube(float);


int main (int argc, char *argv[]) {
	int gameRunning = 1;
	SDL_Event evt;
	
	InitGame();
	while (gameRunning) {
		while (SDL_PollEvent(&evt)) {
			if (evt.type == SDL_QUIT) {
				gameRunning = 0;
			}
			if (evt.type == SDL_KEYDOWN) {
				switch(evt.key.keysym.sym) {
					case SDLK_ESCAPE:
						gameRunning = 0;
						break;
				}
			}
		}
		Display();
		SDL_GL_SwapWindow(win);

	}
	
	
	FinishGame();
	return 0;
}


void InitGame() {
	SDL_Init(SDL_INIT_EVERYTHING);
	win = SDL_CreateWindow("Hello World!", SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, screenWidth, screenHeight, SDL_WINDOW_SHOWN|SDL_WINDOW_OPENGL);
	// SDL_ShowCursor(SDL_DISABLE);
	// SDL_SetRelativeMouseMode(SDL_TRUE);
	
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 2);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 0);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
	
	context = SDL_GL_CreateContext(win);
	// renderer = SDL_CreateRenderer(win, -1, SDL_RENDERER_ACCELERATED|SDL_RENDERER_TARGETTEXTURE);
	
	SDL_GL_SetSwapInterval(1);
	
	InitShaders();
}

void FinishGame () {
	// SDL_DestroyRenderer(renderer);
	SDL_DestroyWindow(win);
	SDL_Quit();
	exit(0);
}


// Simple function to create a shader
void LoadShader(char *Code, int ID)
{
    // Compile the shader code
    glShaderSource  (ID, 1, (const char **)&Code, NULL); 
    glCompileShader (ID);

    // Verify that it worked
    int ShaderStatus;
    glGetShaderiv(ID, GL_COMPILE_STATUS, &ShaderStatus); 

    // Check the compile status
    if (ShaderStatus != GL_TRUE) {
        printf("Error: Failed to compile GLSL program\n");
        int Len = 1024000;
        char Error[1024000];
        glGetShaderInfoLog(ID, 1024000, &Len, Error);
        printf("%s",Error);
        glGetShaderSource(ID, 1024000, &Len, Error);
        printf("%s",Error);
        exit (-1);
    }
}

int InitShaders(void) 
{
	FILE *fp;
	char *VertexShader;
	char *FragmentShader;
	unsigned int flen;

	fp = fopen("outdoor_vertex.glsl", "r");
	if (fp) {
		fseek(fp, 0, SEEK_END);
		flen = ftell(fp);
		fseek(fp, 0, SEEK_SET);
		VertexShader = new char[flen+1];
		memset(VertexShader, 0, flen+1);
		fread(VertexShader, sizeof(char), flen, fp);
		fclose(fp);
	}

	fp = fopen("outdoor_fragment.glsl", "r");
	if (fp) {
		fseek(fp, 0, SEEK_END);
		flen = ftell(fp);
		fseek(fp, 0, SEEK_SET);
		FragmentShader = new char[flen+1];
		memset(FragmentShader, 0, flen+1);
		fread(FragmentShader, sizeof(char), flen, fp);
		fclose(fp);
	}

    // Create 2 shader programs
    Shader[0] = glCreateShader(GL_VERTEX_SHADER);
    Shader[1] = glCreateShader(GL_FRAGMENT_SHADER);

	printf("Loading Vertex Shader (%x)\n", Shader[0]);
	printf("-------------------------\n");
	// printf("%s\n", VertexShader);
    LoadShader((char *)VertexShader, Shader[0]);
    printf("Loading Fragment Shader (%x)\n", Shader[1]);
    printf("-------------------------\n");
    // printf("%s\n", FragmentShader);
    LoadShader((char *)FragmentShader, Shader[1]);

    // Create the prorgam and attach the shaders & attributes
    Program   = glCreateProgram();

    glAttachShader(Program, Shader[0]);
    glAttachShader(Program, Shader[1]);

    glBindAttribLocation(Program, 0, "Position");
    glBindAttribLocation(Program, 1, "Normal");

    // Link
    glLinkProgram(Program);

    // Validate our work thus far
    int ShaderStatus;
    glGetProgramiv(Program, GL_LINK_STATUS, &ShaderStatus); 

    if (ShaderStatus != GL_TRUE) {
        printf("Error: Failed to link GLSL program\n");
        int Len = 1024;
        char Error[1024];
        glGetProgramInfoLog(Program, 1024, &Len, Error);
        printf("%s",Error);
        exit(-1);
    }

    glValidateProgram(Program);

    glGetProgramiv(Program, GL_VALIDATE_STATUS, &ShaderStatus); 

    if (ShaderStatus != GL_TRUE) {
        printf("Error: Failed to validate GLSL program\n");
        exit(-1);
    }
    
    // Enable the program
    glUseProgram                (Program);
    glEnableVertexAttribArray   (0);
    glEnableVertexAttribArray   (1);

    // Setup the Projection matrix
    Persp(Proj, screenWidth, screenHeight, 70.0f, 0.1f, 200.0f);
	// Ortho(Proj, -0.5, 0.05, -0.5*Surface->w/(float)Surface->h, 0.5*Surface->w/(float)Surface->h, 0.1, 100);

    // Retrieve our uniforms
    iProj   = glGetUniformLocation(Program, "Proj");
    iModel  = glGetUniformLocation(Program, "Model");
	iCamera = glGetUniformLocation(Program, "Camera");

    // Basic GL setup
    glClearColor    (0.0, 0.0, 0.0, 1.0);
    glEnable        (GL_CULL_FACE);
    glCullFace      (GL_BACK);
	glEnable		(GL_DEPTH_TEST);

	// clear memory
	delete [] FragmentShader;
	delete [] VertexShader;

    return GL_TRUE;
}

void DrawTourus(float r1, float r2, int resi, int resj) {
	int i, j;
	float theta = 2*PI/(float)resi;
	float phi = 2*PI/(float)resj;
	float x, y, z;
	int offs = 0;

	float *PtData = new float[resi*resj*3];
	float *NormData = new float[resi*resj*3];

	for (j=0;j<resj;j++) {
		for (i=0;i<resi;i++) {
			x = (r1 + r2*cos(phi*j))*cos(theta*i);
			y = r2*sin(phi*j);
			z = (r1 + r2*cos(phi*j))*sin(theta*i);
			PtData[offs+0] = x;
			PtData[offs+1] = y;
			PtData[offs+2] = z;

			x = cos(phi*j)*cos(theta*i);
			y = sin(phi*j);
			z = cos(phi*j)*sin(theta*i);
			NormData[offs+0] = x;
			NormData[offs+1] = y;
			NormData[offs+2] = z;

			offs += 3;
		}
	}

	offs = 0;
	unsigned int *FaceData = new unsigned int [resi*resj*2*3];
	for (j=0;j<resj;j++) {
		for (i=0;i<resi;i++) {
			FaceData[offs+0] = resi*((j+0)%resj)+((i+0)%resi);
			FaceData[offs+1] = resi*((j+1)%resj)+((i+0)%resi);
			FaceData[offs+2] = resi*((j+0)%resj)+((i+1)%resi);
			offs += 3;
			FaceData[offs+0] = resi*((j+0)%resj)+((i+1)%resi);
			FaceData[offs+1] = resi*((j+1)%resj)+((i+0)%resi);
			FaceData[offs+2] = resi*((j+1)%resj)+((i+1)%resi);
			offs += 3;
		}
	}

	glVertexAttribPointer   (0, 3, GL_FLOAT, 0, 0, PtData);
    glVertexAttribPointer   (1, 3, GL_FLOAT, GL_TRUE, 0, NormData);

    glDrawElements          (GL_TRIANGLES, resi*resj*2*3, GL_UNSIGNED_INT, FaceData);

	delete [] FaceData;
	delete [] PtData;
	delete [] NormData;
}

void DrawCube(float CubeSize) {
	// no shared vertices 
	
	float cs = CubeSize/2.0f;
	int i = 0;
	float PtData[4*6*3] = {
		-1.0f, 1.0f, 1.0f,
		-1.0f, -1.0f, 1.0f,
		1.0f, -1.0f, 1.0f,
		1.0f, 1.0f, 1.0f,
		1.0f, 1.0f, 1.0f,
		1.0f, -1.0f, 1.0f,
		1.0f, -1.0f, -1.0f,
		1.0f, 1.0f, -1.0f,
		1.0f, 1.0f, -1.0f,
		1.0f, -1.0f, -1.0f,
		-1.0f, -1.0f, -1.0f,
		-1.0f, 1.0f, -1.0f,
		-1.0f, 1.0f, -1.0f,
		-1.0f, -1.0f, -1.0f,
		-1.0f, -1.0f, 1.0f,
		-1.0f, 1.0f, 1.0f,
		-1.0f, 1.0f, 1.0f,
		1.0f, 1.0f, 1.0f,
		1.0f, 1.0f, -1.0f,
		-1.0f, 1.0f, -1.0f,
		-1.0f, -1.0f, 1.0f,
		-1.0f, -1.0f, -1.0f,
		1.0f, -1.0f, -1.0f,
		1.0f, -1.0f, 1.0f 
	};
	float NormData[4*6*3] = {		
		0.0f, 0.0f, 1.0f,
		0.0f, 0.0f, 1.0f,
		0.0f, 0.0f, 1.0f,
		0.0f, 0.0f, 1.0f,
		1.0f, 0.0f, 0.0f,
		1.0f, 0.0f, 0.0f,
		1.0f, 0.0f, 0.0f,
		1.0f, 0.0f, 0.0f,
		0.0f, 0.0f, -1.0f,
		0.0f, 0.0f, -1.0f,
		0.0f, 0.0f, -1.0f,
		0.0f, 0.0f, -1.0f,
		-1.0f, 0.0f, 0.0f,
		-1.0f, 0.0f, 0.0f,
		-1.0f, 0.0f, 0.0f,
		-1.0f, 0.0f, 0.0f,
		0.0f, 1.0f, 0.0f,
		0.0f, 1.0f, 0.0f,
		0.0f, 1.0f, 0.0f,
		0.0f, 1.0f, 0.0f,
		0.0f, -1.0f, 0.0f,
		0.0f, -1.0f, 0.0f,
		0.0f, -1.0f, 0.0f,
		0.0f, -1.0f, 0.0f 
	};
	unsigned int FaceData[6*2*3] = {
		0, 1, 2,
		0, 2, 3,
		4, 5, 6,
		4, 6, 7,
		8, 9, 10,
		8, 10, 11,
		12, 13, 14,
		12, 14, 15,
		16, 17, 18,
		16, 18, 19,
		20, 21, 22,
		20, 22, 23 
	};
	
	for (i=0;i<6*4*3;i++) {
		PtData[i] *= cs;
	}

	glVertexAttribPointer   (0, 3, GL_FLOAT, 0, 0, PtData);
    glVertexAttribPointer   (1, 3, GL_FLOAT, GL_TRUE, 0, NormData);

    glDrawElements          (GL_TRIANGLES, 6*2*3, GL_UNSIGNED_INT, FaceData);

}


// Main-loop workhorse function for displaying the object
void Display(void)
{
    // Clear the screen
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    float Model[4][4];
    memset(Model, 0, sizeof(Model));
	
	float Camera[4][4];
	memset(Camera, 0, sizeof(Camera));
	
	ccCamera.SetPos(v3d(0, 1.0f*sin(SDL_GetTicks()*0.001f), 1.0f));		// position the camera moving up and down
	ccCamera.SetLookAt(v3d(0, 0, 0));									// look at the origin
	ccCamera.SetUp(v3d(0,1,0));											// set +Y axis as "up"
	ccCamera.SetRoll(0.0f);												// roll viewport with accelerometer angle
	
	ccCamera.GetMatrixArray((float *)Camera);							// get our result matrix for use in GLSL vertex shader

    // Activate GLSL Program
    glUseProgram            (Program);

	// Feed matrices into GLSL
    glUniformMatrix4fv      (iProj, 1, false, (const float *)&Proj[0][0]);
	glUniformMatrix4fv		(iCamera, 1, false, (const float *)&Camera[0][0]);

	// draw tourus
	cmModelviewMatrix.LoadIdentity();
	cmModelviewMatrix.Rotate(SDL_GetTicks() *0.0001f, 0, 0, 1);
	cmModelviewMatrix.GetMatrixArray((float *)Model);
    glUniformMatrix4fv      (iModel, 1, false, (const float *)&Model[0][0]);
	DrawTourus(0.5, 0.1, 32, 16);

	// draw tourus
	cmModelviewMatrix.LoadIdentity();
	cmModelviewMatrix.Rotate(SDL_GetTicks() *0.001f, 0, 1, 0);
	cmModelviewMatrix.Rotate(SDL_GetTicks() *0.0007f, 1, 0, 0);
	cmModelviewMatrix.GetMatrixArray((float *)Model);
    glUniformMatrix4fv      (iModel, 1, false, (const float *)&Model[0][0]);
	// DrawTourus(0.1, 0.025, 32, 16);
	DrawCube(0.25f);
}
